/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "alembic/alembic.pmh"

#define FE_SAA_DEBUG		FALSE

namespace fe
{

BWORD SurfaceAccessibleAbc::load(String a_filename,Real a_frame,
	sp<StringFilterI> a_spFilter)
{
#if FE_SAA_DEBUG
	feLog("SurfaceAccessibleAbc::load file \"%s\"\n",a_filename.c_str());
#endif

	m_frame=a_frame;
	m_spFilter=a_spFilter;

	startJoints();

	if(m_filename!=a_filename)
	{
		m_filename=a_filename;

		delete m_pArchive;
		m_pArchive=NULL;
	}

	if(!m_pArchive)
	{
#if FE_SAA_DEBUG
		feLog("SurfaceAccessibleAbc::load new archive\n");
#endif

		m_pArchive=new Alembic::AbcGeom::IArchive(
				Alembic::AbcCoreHDF5::ReadArchive(),
				a_filename.c_str(),
				Alembic::AbcGeom::ErrorHandler::kQuietNoopPolicy);
	}

	if(!m_pArchive || !m_pArchive->valid())
	{
		return FALSE;
	}

	Alembic::AbcGeom::IObject iObj=m_pArchive->getTop();

	visit(iObj);

	return TRUE;
}

void SurfaceAccessibleAbc::visit(Alembic::AbcGeom::IObject a_iObj)
{
	const String path=a_iObj.getFullName().c_str();
	const U32 childCount=a_iObj.getNumChildren();

#if FE_SAA_DEBUG
	feLog("\nobject \"%s\" %d\n",path.c_str(),childCount);
#endif

	setIdentity(m_refWorld);

	Alembic::AbcGeom::ICompoundProperty properties=a_iObj.getProperties();
	String indent="";
	visitProperties(properties,indent);

	if(Alembic::AbcGeom::IXform::matches(a_iObj.getHeader()))
	{
		Alembic::AbcGeom::IXform ixform(a_iObj,
				Alembic::AbcGeom::kWrapExisting);

		const U32 sampleCount=ixform.getSchema().getNumSamples();

		//* HACK
		U32 sampleIndex=m_frame;
		if(sampleIndex>sampleCount-1)
		{
			sampleIndex=sampleCount-1;
		}

		Alembic::AbcGeom::XformSample sample;
		ixform.getSchema().get(sample,
				Alembic::Abc::ISampleSelector(
				Alembic::Abc::index_t(sampleIndex)));

		Alembic::AbcGeom::M44d m44d=sample.getMatrix();
		double* pRaw=m44d.getValue();

		SpatialTransform xformRelative;
		set(xformRelative,pRaw);

		const Color color(0.0,0.0,1.0);	//* TODO

		const String name=path.basename();
		const String parentName=path.pathname().basename();

		addJoint(name,parentName,xformRelative,m_refWorld,color);
	}

	for(U32 childIndex=0;childIndex<childCount;childIndex++)
	{
		Alembic::AbcGeom::IObject iChild(a_iObj,
				a_iObj.getChildHeader(childIndex).getName());

#if FE_SAA_DEBUG
		feLog("\nof \"%s\" %d/%d \"%s\"\n",path.c_str(),childIndex,childCount,
				iChild.getName().c_str());
#endif

		visit(iChild);
	}

}

void SurfaceAccessibleAbc::visitProperty(
	Alembic::AbcGeom::IScalarProperty a_properties,
	String a_indent)
{
	FEASSERT(!a_properties.isArray());

/*
	const Alembic::AbcCoreAbstract::DataType &dataType=
			a_properties.getDataType();
	const Alembic::Util::uint8_t extent=dataType.getExtent();
	Alembic::Util::Dimensions dimensions(extent);
	Alembic::AbcCoreAbstract::ArraySamplePtr samplePtr=
			Alembic::AbcCoreAbstract::AllocateArraySample(dataType,dimensions);
	const U32 sampleCount=a_properties.getNumSamples();
	for(U32 sampleIndex=0;sampleIndex<sampleCount;++sampleIndex)
	{
		a_properties.get(samplePtr->getData(),
				Alembic::Abc::ISampleSelector(
				Alembic::Abc::index_t(sampleIndex)));
		arraySize=samplePtr->size();
	};
*/

#if FE_SAA_DEBUG
	const String interp=
			a_properties.getMetaData().get("interpretation").c_str();

	std::stringstream dtype;
	dtype << a_properties.getDataType();

	feLog("%sscalar \"%s\" interp '%s' type '%s' samples %d\n",
			a_indent.c_str(),a_properties.getName().c_str(),
			interp.c_str(),dtype.str().c_str(),a_properties.getNumSamples());
#endif
}

void SurfaceAccessibleAbc::visitProperty(
	Alembic::AbcGeom::IArrayProperty a_properties,
	String a_indent)
{
	FEASSERT(a_properties.isArray());

	const String propName=a_properties.getName().c_str();

	std::stringstream dtype;
	dtype << a_properties.getDataType();
	const String dataType=dtype.str().c_str();

#if FE_SAA_DEBUG
	const String interp=
			a_properties.getMetaData().get("interpretation").c_str();

	feLog("%sarray \"%s\" interp '%s' type '%s' samples %d\n",
			a_indent.c_str(),propName.c_str(),
			interp.c_str(),dataType.c_str(),a_properties.getNumSamples());
#endif

	Alembic::AbcCoreAbstract::ArraySamplePtr samplePtr;
	const U32 sampleCount = a_properties.getNumSamples();
	for (U32 sampleIndex=0;sampleIndex<sampleCount;sampleIndex++)
	{
		a_properties.get(samplePtr,
				Alembic::Abc::ISampleSelector(
				Alembic::Abc::index_t(sampleIndex)));
		const U32 arrayCount=samplePtr->size();

		if(dataType=="string")
		{
			std::string* pString=(std::string*)samplePtr->getData();
			const String string=pString[0].c_str();

#if FE_SAA_DEBUG
			for(U32 arrayIndex=0;arrayIndex<arrayCount;arrayIndex++)
			{
				const String nextString=pString[arrayIndex].c_str();
				if(!arrayIndex || string!=nextString)
				{
				feLog("      %d/%d \"%s\"\n",
						arrayIndex,arrayCount,string.c_str());
				}
			}
#endif

			if(propName=="worldreference")
			{
				Real val[16];
				sscanf(string.c_str(),
						"%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f",
						&val[0],&val[1],&val[2],&val[3],
						&val[4],&val[5],&val[6],&val[7],
						&val[8],&val[9],&val[10],&val[11],
						&val[12],&val[13],&val[14],&val[15]);

				set(m_refWorld,val);
			}
		}
	}
}

void SurfaceAccessibleAbc::visitProperty(
	Alembic::AbcGeom::ICompoundProperty a_properties,
	String &a_rIndent)
{
#if FE_SAA_DEBUG
	String schema=a_properties.getMetaData().get("schema").c_str();

	feLog("%scompound \"%s\" schema '%s'\n",
			a_rIndent.c_str(),a_properties.getName().c_str(),schema.c_str());

	String oldIndent=a_rIndent;
	a_rIndent+="  ";
#endif

	visitProperties(a_properties,a_rIndent);

#if FE_SAA_DEBUG
	a_rIndent=oldIndent;
#endif
}

void SurfaceAccessibleAbc::visitProperties(
	Alembic::AbcGeom::ICompoundProperty a_parent,
	String &a_rIndent)
{
#if FE_SAA_DEBUG
	String oldIndent=a_rIndent;
#endif

	for(U32 propIndex=0;propIndex<a_parent.getNumProperties();propIndex++)
	{
		Alembic::AbcGeom::PropertyHeader header=
				a_parent.getPropertyHeader(propIndex);

		if(header.isCompound())
		{
			visitProperty(Alembic::AbcGeom::ICompoundProperty(a_parent,
					header.getName()),a_rIndent);
		}
		else if(header.isScalar())
		{
			visitProperty(Alembic::AbcGeom::IScalarProperty(a_parent,
					header.getName()),a_rIndent);
		}
		else
		{
			FEASSERT(header.isArray());
			visitProperty(Alembic::AbcGeom::IArrayProperty(a_parent,
					header.getName()),a_rIndent);
		}
	}

#if FE_SAA_DEBUG
	a_rIndent=oldIndent;
#endif
}

} // namespace
