/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "maya/maya.pmh"

#define FE_JSO_DEBUG		FALSE

using namespace fe;

void JointSeparateOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<String>("root")="joints";
	catalog<String>("root","label")="Root Node Name";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","implementation")="curve";
	catalog<bool>("Output Surface","array")=true;
}

void JointSeparateOp::handle(Record& a_rSignal)
{
	const String rootName=catalog<String>("root");

#if FE_JSO_DEBUG
	feLog("JointSeparateOp::handle\n");
#endif

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessorI> spInputName;
	if(!access(spInputName,spInputAccessible,e_primitive,"name")) return;

	sp<SurfaceAccessorI> spInputParent;
	if(!access(spInputParent,spInputAccessible,e_primitive,"parent")) return;

	sp<SurfaceAccessorI> spInputAnimX;
	if(!access(spInputAnimX,spInputAccessible,e_primitive,"animX")) return;

	sp<SurfaceAccessorI> spInputAnimY;
	if(!access(spInputAnimY,spInputAccessible,e_primitive,"animY")) return;

	sp<SurfaceAccessorI> spInputAnimZ;
	if(!access(spInputAnimZ,spInputAccessible,e_primitive,"animZ")) return;

	sp<SurfaceAccessorI> spInputAnimT;
	if(!access(spInputAnimT,spInputAccessible,e_primitive,"animT")) return;

	sp<SurfaceAccessibleI> spSurfaceOutput=
			catalog< sp<Component> >("Output Surface");
	if(spSurfaceOutput.isNull())
	{
		catalog<String>("error")+="Inaccessible output;";
		return;
	}

	sp<SurfaceAccessibleMaya> spSurfaceAccessibleMaya=spSurfaceOutput;
	if(spSurfaceAccessibleMaya.isNull())
	{
		catalog<String>("error")+="Incompatible output;";
		return;
	}

	//* TODO is there an easier way to get a node object by name?
	MSelectionList originalList;
	MGlobal::getActiveSelectionList(originalList);

	MGlobal::selectByName(rootName.c_str(),MGlobal::kReplaceList);

	MSelectionList selectionList;
	MGlobal::getActiveSelectionList(selectionList);

	MGlobal::setActiveSelectionList(originalList);

	MStatus status;
	MObject rootObject;
	if(selectionList.length())
	{
		selectionList.getDependNode(0,rootObject);
	}
	if(rootObject.isNull())
	{
		MFnTransform fnTransform;
		rootObject=fnTransform.create(MObject::kNullObj,&status);
		if(!status)
		{
			feLog("JointSeparateOp::handle create root failed: %s\n",
					status.errorString().asChar());
			return;
		}
		fnTransform.setName(rootName.c_str());
	}

	const U32 primitiveCount=spInputName->count();

#if FE_JSO_DEBUG
	feLog("JointSeparateOp::handle primitiveCount %d\n",primitiveCount);
#endif

	std::map<String,MObject> objectMap;

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
#if FE_JSO_DEBUG
		feLog("JointSeparateOp::handle %d/%d\n",primitiveIndex,primitiveCount);
#endif

		const String name=spInputName->string(primitiveIndex);

		const String parentName=spInputParent->string(primitiveIndex);
		MObject parentObject=objectMap[parentName];
		if(parentObject==MObject::kNullObj)
		{
			parentObject=rootObject;
		}

		SpatialTransform xformLocal;
		xformLocal.column(0)=spInputAnimX->spatialVector(primitiveIndex);
		xformLocal.column(1)=spInputAnimY->spatialVector(primitiveIndex);
		xformLocal.column(2)=spInputAnimZ->spatialVector(primitiveIndex);
		xformLocal.column(3)=spInputAnimT->spatialVector(primitiveIndex);

		const SpatialVector localDisplace=xformLocal.translation();

		const SpatialQuaternion quatLocal=xformLocal;

		//* NOTE only need localDisplace and quatLocal
		const MVector mVector(
				localDisplace[0],localDisplace[1],localDisplace[2]);
		const MQuaternion mQuaternion(
				quatLocal[0],quatLocal[1],quatLocal[2],quatLocal[3]);

#if FE_JSO_DEBUG

		feLog("  \"%s\" parent \"%s\"\n",name.c_str(),parentName.c_str());
		feLog("  point %s localDisplace %s\n",
				c_print(xformLocal.translation()),c_print(localDisplace));
		feLog("  quatLocal %s\n",c_print(quatLocal));
#endif

		MObject jointObject;

		MFnDagNode fnDagNode(parentObject);
		const U32 childCount=fnDagNode.childCount(&status);
		for(U32 childIndex=0;childIndex<childCount;childIndex++)
		{
			MObject child=fnDagNode.child(childIndex,&status);
			MFnDagNode dagChild(child);
			if(dagChild.name()==name.c_str())
			{
				jointObject=child;
				break;
			}
		}

		MFnIkJoint fnJoint(jointObject);

		if(jointObject.isNull())
		{
			jointObject=fnJoint.create(parentObject,&status);
			if(!status)
			{
				feLog("JointSeparateOp::handle create failed: %s\n",
						status.errorString().asChar());
				continue;
			}

			fnJoint.setName(name.c_str());
		}

		objectMap[name]=jointObject;

		fnJoint.setTranslation(mVector,MSpace::kTransform);
		fnJoint.setRotation(mQuaternion,MSpace::kTransform);
	}

#if FE_JSO_DEBUG
	feLog("JointSeparateOp::handle done\n");
#endif
}
